(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('vue')) :
  typeof define === 'function' && define.amd ? define(['exports', 'vue'], factory) :
  (global = global || self, factory(global.AsyncManager = {}, global.Vue));
}(this, function (exports, Vue) { 'use strict';

  Vue = Vue && Vue.hasOwnProperty('default') ? Vue['default'] : Vue;

  var currentInstance = null;
  function setCurrentInstance(ins) {
      currentInstance = ins;
  }
  var suspenseInstanceStack = [];
  var currentSuspenseInstance;
  function pushSuspenseInstance(ins) {
      currentSuspenseInstance = ins;
      suspenseInstanceStack.push(ins);
  }
  function popSuspenseInstance() {
      suspenseInstanceStack.pop();
      return (currentSuspenseInstance =
          suspenseInstanceStack[suspenseInstanceStack.length - 1]);
  }

  var RESOLVED = 'resolved';
  var REJECTED = 'rejected';
  var del = function (af, error) {
      var suspIns = af.suspenseInstance;
      if (!suspIns) {
          {
              console.error('No Suspense instance');
          }
          return;
      }
      var asyncFactorys = suspIns.asyncFactorys;
      if (error) {
          suspIns.$emit(REJECTED, error);
          return;
      }
      asyncFactorys.delete(af);
      if (asyncFactorys.size === 0) {
          suspIns.$emit(RESOLVED);
      }
  };
  var add = function (af) {
      var suspIns = currentSuspenseInstance || af.suspenseInstance;
      if (!suspIns) {
          {
              console.error('No Suspense instance');
          }
          return;
      }
      var asyncFactorys = suspIns.asyncFactorys || (suspIns.asyncFactorys = new Set());
      if (suspIns.resolved) {
          suspIns.resolved = false;
          suspIns.setupLoading();
      }
      asyncFactorys.add(af);
  };
  var has = function (af) {
      var suspIns = currentSuspenseInstance || af.suspenseInstance;
      if (!suspIns) {
          {
              console.error('No Suspense instance');
          }
          return;
      }
      return suspIns.asyncFactorys && suspIns.asyncFactorys.has(af);
  };
  var COMPONENT_NAME = 'VueSuspense';
  var Suspense = {
      name: COMPONENT_NAME,
      data: function () {
          return {
              resolved: false,
              displayLoading: false
          };
      },
      props: {
          delay: {
              type: Number,
              default: 0
          }
      },
      methods: {
          setupLoading: function () {
              var _this = this;
              // Setting loading status
              if (this.delay > 0) {
                  this._timer = setTimeout(function () {
                      _this.displayLoading = true;
                  }, this.delay);
              }
              else {
                  this.displayLoading = true;
              }
          },
          destroyLoading: function () {
              if (this._timer) {
                  clearTimeout(this._timer);
                  this._timer = null;
              }
              this.displayLoading = false;
          }
      },
      created: function () {
          var _this = this;
          pushSuspenseInstance(this);
          // `this.promiser` is for test cases
          this.promiser = new Promise(function (resolve, reject) {
              _this.$on(RESOLVED, function () {
                  _this.destroyLoading();
                  _this.resolved = true;
                  resolve();
              });
              _this.$on(REJECTED, function (err) {
                  _this.destroyLoading();
                  _this.rejected = true;
                  reject(err);
              });
          });
          this.setupLoading();
      },
      mounted: function () {
          if (!this.asyncFactorys) {
              // This means that there are no lazy components or resource.read()
              // in the child components of the Suspense component,
              // set to resolved to update rendering.
              // Warning: If the content wrapped by the Suspense component is static, the static content will be rendered twice.
              this.$emit(RESOLVED);
          }
      },
      updated: function () {
          /**
           * If the asynchronous call has not been resolved,
           * the reason for triggering the update is to do something else,
           * such as: display loading, etc., so popSuspenseInstance() should not be called.
           */
          if (!this.resolved)
              return;
          popSuspenseInstance();
      },
      render: function (h) {
          var isVisible = this.$options.suspense.mode === 'visible';
          var emptyVNode = this._e();
          var fallback = this.displayLoading
              ? this.$slots.fallback || [emptyVNode]
              : [emptyVNode];
          // The `tree` is the real content to be rendered
          var tree = this.$slots.default || [emptyVNode];
          var rendered;
          if (this.rejected && this.$slots.error) {
              rendered = createWrapper(h, this.$slots.error);
          }
          else {
              rendered = this.resolved
                  ? isVisible
                      ? createWrapper(h, tree)
                      : createWrapper(h, [
                          // We need to render the tree, but we should not show the rendered content.
                          createHiddenWrapper(h, tree, true)
                      ])
                  : isVisible
                      ? createWrapper(h, tree.concat(fallback))
                      : createWrapper(h, [
                          // We need to render the tree, but we should not show the rendered content.
                          createHiddenWrapper(h, tree, false),
                          fallback
                      ]);
          }
          return rendered;
      }
  };
  function createWrapper(h, children) {
      return h('div', {
          class: { 'vue-suspense-wrapper': true }
      }, children);
  }
  function createHiddenWrapper(h, tree, display) {
      return h('div', {
          style: { display: display ? 'block' : 'none' },
          class: { 'vue-suspense-hidden-wrapper': true }
      }, tree);
  }

  function findSuspenseInstance(ins) {
      var current = ins.$parent;
      while (current) {
          if (current.$options.name === COMPONENT_NAME) {
              return current;
              break;
          }
          else {
              current = current.$parent;
          }
      }
      return;
  }

  function lazy(asyncFactory, props) {
      return {
          name: 'SSLazy',
          props: props || [],
          created: function () {
              var _this = this;
              asyncFactory.suspenseInstance =
                  currentSuspenseInstance || findSuspenseInstance(this);
              if (has(asyncFactory))
                  return;
              add(asyncFactory);
              if (asyncFactory.resolved) {
                  asyncFactory.$$waiter.then(function () {
                      del(asyncFactory);
                  });
                  return;
              }
              var promise = asyncFactory();
              asyncFactory.$$waiter = promise;
              promise
                  .then(function (C) {
                  // Compatible ES Module
                  if (C.__esModule && C.default) {
                      C = C.default;
                  }
                  asyncFactory.resolved = C;
                  // Trigger update
                  _this.$forceUpdate();
              })
                  .catch(function (err) {
                  {
                      console.error(err);
                  }
                  del(asyncFactory, err);
              });
          },
          updated: function () {
              del(asyncFactory);
          },
          render: function (h) {
              var _this = this;
              // Fix context
              var slots = Object.keys(this.$slots)
                  .reduce(function (arr, key) { return arr.concat(_this.$slots[key] || []); }, [])
                  .map(function (vnode) {
                  vnode.context = _this._self;
                  return vnode;
              });
              return asyncFactory.resolved
                  ? h(asyncFactory.resolved, {
                      on: this.$listeners,
                      // If there is no props definition, fall back to `this.$attrs`
                      props: props ? this.$props : this.$attrs,
                      // Passthrough scopedSlots
                      scopedSlots: this.$scopedSlots,
                      attrs: this.$attrs
                  }, slots)
                  : this._e();
          }
      };
  }

  Vue.mixin({
      created: function () {
          setCurrentInstance(this);
      }
  });
  function observable(data) {
      if (Vue.observable) {
          return Vue.observable(data);
      }
      return new Vue({
          data: function () {
              return data;
          }
      }).$data;
  }
  function createResource(fetchFactory, options) {
      var $res = observable({
          $$result: null,
          $$error: null,
          $$loading: false
      });
      var ins = currentInstance;
      fetchFactory.suspenseInstance = currentSuspenseInstance
          ? currentSuspenseInstance
          : findSuspenseInstance(ins);
      var hasSuspenseInstance = !!fetchFactory.suspenseInstance;
      var resourceManager = {
          read: function (input) {
              if (options && options.prevent && $res.$$loading) {
                  return $res.$$promiser;
              }
              $res.$$loading = true;
              // Because we don't need caching, this is just a unique identifier,
              // and each call to .read() is a completely new request.
              var uniqueWrapFactory = function (i) {
                  return fetchFactory(i);
              };
              if (hasSuspenseInstance) {
                  // Establish a relationship between the fetchFactory and the current component instance
                  uniqueWrapFactory.suspenseInstance = fetchFactory.suspenseInstance;
                  add(uniqueWrapFactory);
              }
              // Start fetching asynchronous data
              var promise = ($res.$$promiser = uniqueWrapFactory(input));
              promise
                  .then(function (res) {
                  // Trigger update
                  $res.$$result = res;
                  $res.$$loading = false;
                  if (hasSuspenseInstance)
                      del(uniqueWrapFactory);
              })
                  .catch(function (err) {
                  {
                      console.error(err);
                  }
                  $res.$$error = err;
                  $res.$$loading = false;
                  if (hasSuspenseInstance)
                      del(uniqueWrapFactory, err);
              });
              return promise;
          },
          get $result() {
              return $res.$$result;
          },
          set $result(val) {
              $res.$$result = val;
          },
          get $error() {
              return $res.$$error;
          },
          set $error(val) {
              $res.$$error = val;
          },
          get $loading() {
              return $res.$$loading;
          },
          fork: function () {
              return createResource(function (i) { return fetchFactory(i); });
          }
      };
      return resourceManager;
  }

  function install(Vue, options) {
      Vue.component('Suspense', Suspense);
      var opts = Object.assign({
          mode: 'visible'
      }, options);
      Vue.setSuspenseOptions = function (options) {
          Object.assign(opts, options);
      };
      Vue.mixin({
          created: function () {
              if (this.$options.name === COMPONENT_NAME) {
                  this.$options.suspense = opts;
              }
          }
      });
  }
  if (window && window.Vue) {
      window.Vue.use(install);
  }

  exports.Suspense = Suspense;
  exports.createResource = createResource;
  exports.default = install;
  exports.lazy = lazy;

  Object.defineProperty(exports, '__esModule', { value: true });

}));
